/* ============================================================
 * File  : videowrapper.cpp
 * Author: Stephane Gibault
 * Date  : Sun Feb 4 2007
 * Description :
 *
 *
 *  (C) 2007 by Stephane Gibault
 *
 * This program is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software Foundation;
 * either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * ============================================================ */
#include <QtGui>
#include "videowrapper.h"

videowrapper::videowrapper()
{
    av_register_all();
}

void videowrapper::delete_videowrapper()
{
    av_free_packet(&packet);
    //av_free(pFrame);
    //av_free(pFrameRGB);

    //av_free_packet(&packet);
    avcodec_close(pCodecCtx);
    av_close_input_file(pFormatCtx);
}

int videowrapper::load_video(QString filename)
{
    // *** OUVERTURE DU FICHIER VIDEO ***
    #ifdef avformat_open_input
        if (avformat_open_input(&pFormatCtx,filename.toAscii().data(),NULL,NULL)!=0)
        {
    #else
        if (av_open_input_file(&pFormatCtx,filename.toAscii().data(),NULL,0,NULL)!=0)
        {
    #endif
        qDebug() << "CAN'T OPEN VIDEO FILE : " << filename;
        return 1;
    }

    // *** RECUPERATION DES INFORMATIONS DU MEDIA ***
    if(av_find_stream_info(pFormatCtx)<0)
    {
        qDebug() << "CAN'T READ INFORMATION FROM VIDEO FILE : " << filename;
        return 1;
    }

    // *** RECHERCHE STREAM VIDEO ***

    videoStream=-1;
    for( i=0; i < pFormatCtx->nb_streams; i++ )
    {
        if(pFormatCtx->streams[i]->codec->codec_type==AVMEDIA_TYPE_VIDEO)
        {
            videoStream=i;
            break;
        }
    }

    if( videoStream == -1 )
    {
      qDebug() << "CAN'T FIND VIDEO STREAM : " << filename;
      return 1;
    }

    pCodecCtx=pFormatCtx->streams[videoStream]->codec;

    pCodec=avcodec_find_decoder(pCodecCtx->codec_id);
    if( pCodec==NULL )
    {
        qDebug() << "UNSUPPORTED CODEC";
        return 1;
    }

    // *** OUVERTURE DU CODEC ***
    if(avcodec_open(pCodecCtx, pCodec)<0)
    {
        qDebug() << "CAN'T FIND VIDEO CODEC";
        return 1;
    }

    img_convert_ctx = sws_getContext(pCodecCtx->width,
                    pCodecCtx->height, pCodecCtx->pix_fmt, pCodecCtx->width,
                    pCodecCtx->height, PIX_FMT_RGB24, SWS_BICUBIC, NULL,
                    NULL, NULL);

    // *** ON DETERMINE LA TAILLE DU BUFFER ***
    numBytes=avpicture_get_size(PIX_FMT_RGB24, pCodecCtx->width,
                                pCodecCtx->height);
    buffer=(uint8_t *)av_malloc(numBytes*sizeof(uint8_t));
    old_frame = -1;
    old_result = 0;

    return 0;
}

QImage videowrapper::read_frame(int64_t num_frame)
{
    if ( num_frame != old_frame )
    {
        // *** ALLOCATE VIDEO FRAME ***
        pFrameRGB=avcodec_alloc_frame();
        pFrame=avcodec_alloc_frame();

        avpicture_fill((AVPicture *)pFrameRGB, buffer, PIX_FMT_RGB24,
                        pCodecCtx->width, pCodecCtx->height);

        QImage img(pCodecCtx->width,pCodecCtx->height,QImage::Format_RGB888);

        posms = (float)((float)num_frame/(float)25)*(float)1000;
        DesiredFrameNumber = av_rescale(posms,pFormatCtx->streams[videoStream]->time_base.den,pFormatCtx->streams[videoStream]->time_base.num);
        DesiredFrameNumber/=1000;

        if ( old_result == 0 )
        {
            if ( num_frame < old_frame )
            {
                avformat_seek_file(pFormatCtx,videoStream,DesiredFrameNumber-1,DesiredFrameNumber,DesiredFrameNumber+1,AVSEEK_FLAG_FRAME);
                compte_frame = num_frame-1;
            }
            else
            {
                if ( num_frame < old_frame+25 )
                {
                    compte_frame = old_frame;
                }
                else
                {
                    avformat_seek_file(pFormatCtx,videoStream,DesiredFrameNumber-1,DesiredFrameNumber,DesiredFrameNumber+1,AVSEEK_FLAG_FRAME);
                    compte_frame = num_frame-1;
                }
            }
        }
        else
        {
            if ( num_frame < old_frame )
            {
                av_seek_frame(pFormatCtx,videoStream,0, AVSEEK_FLAG_FRAME);
                compte_frame = 0;
            }
        }

        while(av_read_frame(pFormatCtx, &packet)>=0 && compte_frame < 9999)
        {
            if(packet.stream_index == videoStream)
            {
                avcodec_decode_video2(pCodecCtx, pFrame, &frameFinished, &packet);

                if(frameFinished)
                {
                    if ( compte_frame == num_frame )
                    {
                        sws_scale(img_convert_ctx, pFrame->data, pFrame->linesize, 0, pCodecCtx->height, pFrameRGB->data, pFrameRGB->linesize);

                        for(int y=0;y<pCodecCtx->height;y++)
                        {
                            memcpy(img.scanLine(y),pFrameRGB->data[0]+y*pFrameRGB->linesize[0],pCodecCtx->width*3);
                            oldimg = img;
                        }
                        compte_frame = 9999;
                    }
                    else
                    {
                        av_free_packet(&packet);
                    }
                    compte_frame++;
                }
            }
        }

        if ( compte_frame < 9999 )
        {
            img = oldimg;
            old_result = 1;
        }
        else
        {
            old_result = 0;
        }

        av_free_packet(&packet);
        av_free(pFrame);
        av_free(pFrameRGB);

        old_frame = num_frame;

        return img;
    }
    else
    {
        return oldimg;
    }
}
